from sqlalchemy import Column, Integer, String, UniqueConstraint, DateTime
from sqlalchemy.orm import relationship
from models.base import Base, Session, logger
from datetime import datetime

class Author(Base):
    __tablename__ = 'authors'

    id = Column(Integer, primary_key=True)
    name = Column(String(100))
    number = Column(String(50))
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    last_updated = Column(DateTime, default=datetime.utcnow, onupdate=datetime.utcnow)

    messages = relationship("Message", back_populates="author")

    @property
    def message_count(self):
        return len(self.messages) if self.messages is not None else 0

    __table_args__ = (
        # Add unique constraint on name and number combination
        UniqueConstraint('name', 'number', name='uix_author_name_number'),
    )

    @classmethod
    def get_or_create(cls, name, number=None):
        if not name or name.strip() == "":
            name = "Unknown Author"
        else:
            name = name.strip()

        if number:
            number = str(number).strip()

        # Create a new session for this operation
        session = Session()

        try:
            # Try to get existing author
            query = session.query(cls).filter(cls.name == name)
            if number is None:
                query = query.filter(cls.number.is_(None))
            else:
                query = query.filter(cls.number == number)

            author = query.first()

            if author:
                author_id = author.id
                # Update number if needed
                if number and (not author.number or author.number.strip() == ''):
                    author.number = number
                    session.commit()
                return author_id

            # Create new author if not found
            new_author = cls(name=name, number=number)
            session.add(new_author)
            session.commit()
            return new_author.id

        except Exception as e:
            session.rollback()
            logger.error(f"Error in get_or_create author: {str(e)}")

            # Only attempt default author creation if not already trying it
            if name != "Unknown Author" or number is not None:
                session.close()  # Close current session before recursive call
                try:
                    return cls.get_or_create("Unknown Author", None)
                except Exception as inner_e:
                    logger.error(f"Failed to create default author: {str(inner_e)}")
                    raise
            raise
        finally:
            session.close()

    def to_dict(self, include_messages=False):
        result = {
            'id': self.id,
            'name': self.name,
            'number': self.number,
            'message_count': getattr(self, 'message_count', 0),
            'created_at': self.created_at.isoformat() if self.created_at else None,
            'last_updated': self.last_updated.isoformat() if self.last_updated else None,
            'latest_timestamp': None
        }

        # Only include messages if specifically requested and if we're in an active session
        if include_messages:
            result['messages'] = []

            # Check if we have access to messages and if we're in an active session
            try:
                # Try to safely access messages
                if hasattr(self, 'messages') and self.messages is not None:
                    # Get message timestamps safely
                    timestamps = []
                    message_dicts = []

                    for m in self.messages[:10]:
                        try:
                            if hasattr(m, 'timestamp') and m.timestamp:
                                timestamps.append(m.timestamp)

                            if include_messages:
                                message_dicts.append({
                                    'id': getattr(m, 'id', None),
                                    'caption': getattr(m, 'caption', None),
                                    'timestamp': m.timestamp.isoformat() if hasattr(m, 'timestamp') and m.timestamp else None
                                })
                        except Exception as msg_err:
                            logger.warning(f"Error processing individual message: {str(msg_err)}")

                    # Set latest timestamp if we have any
                    if timestamps:
                        result['latest_timestamp'] = max(timestamps).isoformat()

                    # Set messages if requested
                    if include_messages:
                        result['messages'] = message_dicts
            except Exception as e:
                logger.error(f"Error safely accessing messages in to_dict: {str(e)}")
                # Don't let message processing errors break the whole response
                result['latest_timestamp'] = None
                result['messages'] = []

        return result

    @classmethod
    def create(cls, data):
        """Create a new author"""
        session = Session()
        try:
            # Ensure timestamps are set
            current_time = datetime.utcnow()
            if 'created_at' not in data:
                data['created_at'] = current_time
            if 'last_updated' not in data:
                data['last_updated'] = current_time

            logger.info(f"Creating new author with data: {data}")
            author = cls(**data)
            session.add(author)
            session.commit()
            logger.info(f"Successfully created author with ID: {author.id}")
            return author
        except Exception as e:
            session.rollback()
            logger.error(f"Error creating author: {str(e)}")
            raise
        finally:
            session.close()

    @classmethod
    def update(cls, author_id, data):
        """Update existing author"""
        session = Session()
        try:
            author = session.query(cls).get(author_id)
            if not author:
                logger.warning(f"Author not found with ID: {author_id}")
                return None

            logger.info(f"Found author to update: {author.id}")

            # Create a dictionary of updates
            updates = {}
            updateable_fields = ['name', 'number']
            for field in updateable_fields:
                if field in data:
                    old_value = getattr(author, field)
                    new_value = data[field]
                    logger.info(f"Updating {field} from {old_value} to {new_value}")
                    updates[field] = new_value

            # Note: last_updated timestamp will be automatically updated by SQLAlchemy
            # due to the onupdate=datetime.utcnow parameter in the model definition

            # Update the author using the session's query
            session.query(cls).filter_by(id=author.id).update(updates)
            session.commit()

            logger.info(f"Updated author {author.id}")

            # Fetch the updated author from the database
            updated_author = session.query(cls).get(author.id)

            # Create a dictionary representation before closing the session
            # This avoids lazy loading issues after the session is closed
            author_dict = {
                'id': updated_author.id,
                'name': updated_author.name,
                'number': updated_author.number,
                'message_count': getattr(updated_author, 'message_count', 0),
                'created_at': updated_author.created_at.isoformat() if updated_author.created_at else None,
                'last_updated': updated_author.last_updated.isoformat() if updated_author.last_updated else None
            }

            logger.info(f"Successfully updated author: {author_id}")

            # Create a new instance with the dictionary data
            # This avoids the session binding issue
            result_author = cls()
            for key, value in author_dict.items():
                setattr(result_author, key, value)

            return result_author
        except Exception as e:
            session.rollback()
            logger.error(f"Error updating author: {str(e)}")
            raise
        finally:
            session.close()

    @classmethod
    def delete(cls, author_id):
        """Delete an author"""
        session = Session()
        try:
            author = session.query(cls).get(author_id)
            if author:
                session.delete(author)
                session.commit()
                logger.info(f"Successfully deleted author: {author_id}")
                return True
            logger.warning(f"Author not found for deletion: {author_id}")
            return False
        except Exception as e:
            session.rollback()
            logger.error(f"Error deleting author: {str(e)}")
            raise
        finally:
            session.close()

    @classmethod
    def get_all(cls, page=1, per_page=20):
        """Get all authors with pagination"""
        session = Session()
        try:
            query = session.query(cls)
            total = query.count()
            authors = query.order_by(cls.last_updated.desc())\
                        .offset((page - 1) * per_page)\
                        .limit(per_page)\
                        .all()

            return authors, total
        finally:
            session.close()

    @classmethod
    def search(cls, query_text, page=1, per_page=20):
        """Search authors"""
        session = Session()
        try:
            query = session.query(cls).filter(
                cls.name.ilike(f"%{query_text}%") |
                cls.number.ilike(f"%{query_text}%")
            )

            total = query.count()
            authors = query.order_by(cls.last_updated.desc())\
                        .offset((page - 1) * per_page)\
                        .limit(per_page)\
                        .all()

            return authors, total
        finally:
            session.close()
